from __future__ import annotations

import itertools
import sys
import warnings
from collections.abc import Awaitable, Coroutine, Generator, Iterable, Sequence
from contextlib import asynccontextmanager
from typing import Any, Callable, TypeVar

import anyio

from .exceptions import ParamsError

if sys.version_info >= (3, 11):  # pragma: no cover
    from typing import TypeVarTuple, Unpack
else:
    from exceptiongroup import ExceptionGroup  # pragma: no cover
    from typing_extensions import TypeVarTuple, Unpack  # pragma: no cover


T_Retval = TypeVar("T_Retval")
PosArgsT = TypeVarTuple("PosArgsT")
AsyncFunc = Callable[..., Coroutine]


def ensure_afunc(
    coro: Coroutine[None, None, T_Retval] | Callable[..., Awaitable[T_Retval]],
) -> Callable[..., Awaitable[T_Retval]]:
    """Wrap coroutine to be async function"""
    if callable(coro):
        return coro

    async def do_await():
        return await coro

    return do_await


def run_async(
    coro: Coroutine[None, None, T_Retval] | Callable[..., Awaitable[T_Retval]],
) -> T_Retval:
    """Deprecated! Usage `asynctor.run` instead.

    Usage::
        >>> async def afunc(n=1):
        ...     return n
        ...
        >>> run_async(afunc)  # get the same result as: await afunc()
        1
        >>> run_async(afunc(2))  # get the same result as: await afunc(2)
        2
    """
    return anyio.run(ensure_afunc(coro))


def run(
    func: (Coroutine[None, None, T_Retval] | Callable[[Unpack[PosArgsT]], Awaitable[T_Retval]]),
    *args: Unpack[PosArgsT],
    backend: str = "asyncio",
    backend_options: dict[str, Any] | None = None,
) -> T_Retval:
    """Combine `asyncio.run` and `anyio.run`

    :param func: async function or coroutine.
    :param args: arguments that will pass to `func` if it's a function.
    :param backend: should be 'asyncio' or 'trio'.
    :param backend_options: will pass to `anyio.run`.

    Usage::

    ... code-block:: python3

        from functools import partial

        async def foo(a, b, *, c=3):
            return a, b, c

        result_asyncio_format = run(foo(1, 2, c=0))
        result_anyio_format = run(partial(foo, c=0), 1, 2)
        assert result_asyncio_format == result_anyio_format

    """
    if not callable(func):

        async def do_await() -> T_Retval:
            return await func

        return anyio.run(do_await, backend=backend, backend_options=backend_options)
    return anyio.run(func, *args, backend=backend, backend_options=backend_options)


async def bulk_gather(
    coros: Sequence[Coroutine] | Generator[Coroutine, None, None],
    batch_size=0,
    wait_last=False,
    raises=True,
    *,
    limit: int | None = None,
) -> tuple:
    """
    Similar like `asyncio.gather`, but support the `batch_size` parameter.
    If batch_size is not zero, running tasks will CapacityLimiter({batch_size}).

    :param coros: Coroutines
    :param batch_size: running tasks limit number, set 0 to be unlimit.
    :param wait_last: if True, wait last bulk tasks to complete then start new task group,
        else use anyio.CapacityLimiter to limit task number.
    :param raises: if True, raise Exception when coroutine failed, else return None.
    :param limit: (deprecated) only leave it here to compare with old version.
    """
    try:
        total = len(coros)  # type:ignore[arg-type]
    except TypeError:  # if coros is generator
        total = 0
    else:
        if total == 0:
            return ()
    results = [None] * total

    async def runner(_i: int, _coro: Coroutine) -> None:
        results[_i] = await _coro

    async def limited_runner(_i, _coro, _limiter) -> None:
        async with _limiter:
            results[_i] = await _coro

    try:
        if limit is not None:
            if batch_size:
                if batch_size != limit:
                    raise ParamsError(f"Conflict value with {limit=} & {batch_size=}")
                warnings.warn(
                    "`limit` is deprecated, use `batch_size` only.",
                    DeprecationWarning,
                    stacklevel=2,
                )
            else:
                batch_size = limit
        if batch_size:
            if wait_last:
                if total == 0:
                    todos: list[tuple[int, Coroutine]] = []
                    for count, args in zip(itertools.count(), enumerate(coros)):
                        todos.append(args)
                        results.append(None)
                        if count % batch_size == 0:
                            await map_group(runner, todos)
                            todos.clear()
                    if todos:
                        await map_group(runner, todos)
                else:
                    for start in range(0, total, batch_size):
                        coros_slice = itertools.islice(coros, start, start + batch_size)
                        args_items = ((start + i, coro) for i, coro in enumerate(coros_slice))
                        await map_group(runner, args_items)
            else:
                limiter = anyio.CapacityLimiter(batch_size)
                todo_args = ((*item, limiter) for item in enumerate(coros))
                if total == 0:
                    await map_group(limited_runner, todo_args, results)
                else:
                    await map_group(limited_runner, todo_args)
        else:
            if total == 0:
                await map_group(runner, enumerate(coros), results)
            else:
                await map_group(runner, enumerate(coros))
    except ExceptionGroup as e:
        if raises:
            raise e.exceptions[0] from e

    return tuple(results)


async def map_group(func: AsyncFunc, todos: Iterable, results: list | None = None) -> None:
    """
    The `map_group` function asynchronously executes a given function for each set of arguments in a
    provided iterable using `anyio.create_task_group()`.

    :param func: The `func` parameter is an asynchronous function that will be applied to
        each item in the `todos` iterable
    :param todos: The `todos` parameter in the `map_group` function represents an iterable
        containing the arguments that will be passed to the `func` function for each task
        that is started in the task group. Each element in `todos` is a set of arguments
        that will be unpacked and passed to the `func`.
    :param results: The `results` parameter in the `map_group` function is a list that
        stores the results of the asynchronous function calls.
        If the `results` parameter is provided with an initial list,
        the results of each function call will be appended to this list.
        If `results` is None or not provided, tasks will run without any return.
    """
    async with anyio.create_task_group() as tg:
        for args in todos:
            if results is not None:
                results.append(None)
            tg.start_soon(func, *args)


async def gather(*coros: Coroutine) -> tuple:
    """Similar like asyncio.gather"""
    return await bulk_gather(coros)


@asynccontextmanager
async def start_tasks(coro: Coroutine | Callable, *more: Coroutine | Callable):
    """Make it easy to convert asyncio.create_task

    Usage:

    ... code-block:: python3

        async def startup():
            # cost a long time to do sth async
            await anyio.sleep(1000)

        @contextlib.asynccontextmanager
        async def lifespan(app):
            async with start_tasks(startup()):
                yield
    """
    async with anyio.create_task_group() as tg:
        with anyio.CancelScope(shield=True):
            tg.start_soon(ensure_afunc(coro))
            for c in more:
                tg.start_soon(ensure_afunc(c))
            try:
                yield
            finally:
                tg.cancel_scope.cancel()


async def wait_for(coro: Coroutine[None, None, T_Retval], timeout: int | float) -> T_Retval:
    """Similar like asyncio.wait_for"""
    with anyio.fail_after(timeout):
        return await coro



from __future__ import annotations

import concurrent.futures
import functools
import sys
from collections.abc import Callable
from contextlib import AbstractContextManager
from functools import cached_property
from threading import Thread
from typing import Annotated, Any, TypeVar

if sys.version_info >= (3, 11):  # pragma: no cover
    from typing import ParamSpec, Self
else:  # pragma: no cover
    from typing_extensions import ParamSpec, Self

T_Retval = TypeVar("T_Retval")
T_ParamSpec = ParamSpec("T_ParamSpec")
FuncResults = Annotated[list, "The return value of each funtion or the exception that it raises"]


class StoredThread(Thread):
    def run(self):
        """Use a new attribute `_result` to stored target result or exception."""
        try:
            self._result = self._target(*self._args, **self._kwargs)  # type:ignore[attr-defined]
        except Exception as e:
            self._result = e


class ThreadGroup(AbstractContextManager):
    """
    Make it easy to wait for threads to complete, and better support for autocompletion.

    Use it like this:

    ```Python
    def do_work(arg1, arg2, kwarg1="", kwarg2="") -> str:
        # Do work

    with ThreadGroup() as tg:
        for _ in range(10):
            tg.soonify(do_work)("spam", "ham", kwarg1="a", kwarg2="b")

    print(f'All {len(tg.threads)} threads completed.')
    ```

    ## Arguments

    `max_workers`: if != 0, use `concurrent.futures.ThreadPoolExecutor(max_workers)`
    `timeout`: floating point number that specifying a timeout for the operation in seconds.
    """

    def __init__(self, max_workers: int | None = 0, timeout: float | None = None):
        self._threads: list[StoredThread] = []
        self._results: list[Any] = []
        self._timeout = timeout
        self._max_workers = max_workers
        self._executor: concurrent.futures.ThreadPoolExecutor | None = None
        self._future_idx: dict[concurrent.futures.Future, int] = {}

    @property
    def results(self) -> FuncResults:
        return self._results

    @cached_property
    def use_pool(self) -> bool:
        """Whether use ThreadPoolExecutor"""
        return self._max_workers != 0

    def __enter__(self) -> Self:
        if self.use_pool:
            self._executor = concurrent.futures.ThreadPoolExecutor(
                max_workers=self._max_workers
            ).__enter__()
        return self

    def soonify(self, func: Callable[T_ParamSpec, Any]) -> Callable[T_ParamSpec, None]:
        """
        Create and start a Thread instance then add it to this group.

        Use it like this:

        ```Python
        with ThreadGroup() as tg:
            def do_work(arg1, arg2, kwarg1="", kwarg2="") -> str:
                # Do work

            tg.soonify(do_work)("spam", "ham", kwarg1="a", kwarg2="b")

        ```
        """

        @functools.wraps(func)
        def runner(*args: T_ParamSpec.args, **kwargs: T_ParamSpec.kwargs) -> None:
            if self._executor is not None:
                fut = self._executor.submit(func, *args, **kwargs)
                self._future_idx[fut] = len(self._future_idx)
            else:
                t = StoredThread(target=func, args=args, kwargs=kwargs)
                t.start()
                self._threads.append(t)

        return runner

    def __exit__(self, *args, **kwargs):
        if fs := self._future_idx:
            self._results = [None] * len(fs)
            for future in concurrent.futures.as_completed(fs, timeout=self._timeout):
                idx = fs[future]
                try:
                    res = future.result()
                except Exception as exc:
                    res = exc
                self._results[idx] = res
        else:
            for t in self._threads:
                t.join(timeout=self._timeout)
            for t in self._threads:
                if t.is_alive():
                    fn, gs, kw = t._target, t._args, t._kwargs  # type:ignore[attr-defined]
                    if len(msg := f"{fn}(*{gs!r}, **{kw!r})") > 50:
                        msg = msg[:47] + "..."
                    res = TimeoutError(msg)
                else:
                    res = t._result
                self._results.append(res)


def _test() -> None:  # pragma: no cover
    import time

    def saying(msg="haha", loop=5, ident=0) -> float:
        print(msg)
        total = 0
        for i in range(loop):
            print(f"Sleeping {ident}: {i}/10 seconds")
            time.sleep(i / 10)
            total += i
        return round(total / 10, 1)

    with ThreadGroup() as tg:
        for i in range(10):
            tg.soonify(saying)(f"t{i + 1}", ident=i, loop=((i + 1) * 2 % 5 + 1))
    print(tg.results)
    print("Done.")


if __name__ == "__main__":
    _test()



from __future__ import annotations

from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager
from typing import Annotated, Any

from fastapi import Depends, FastAPI, Request
from fastapi.routing import _merge_lifespan_context

from ..client import AsyncRedis


def register_aioredis(
    app: FastAPI,
    check_connection: bool = True,
    **kwargs: Annotated[Any, "Kwargs that will pass to `redis.asyncio.Redis.__init__`"],
) -> None:
    """
    Register redis to fastapi application

    :param app: the fastapi application instance
    :param check_connection: whether check redis server pingable
    :param kwargs: such like: host, port, etc.
    """

    @asynccontextmanager
    async def redis_lifespan(app_instance: FastAPI) -> AsyncGenerator[None]:
        async with AsyncRedis(app_instance, **kwargs):
            yield

    original_lifespan = app.router.lifespan_context
    app.router.lifespan_context = _merge_lifespan_context(redis_lifespan, original_lifespan)


async def get_redis_client(request: Request) -> AsyncRedis:
    return AsyncRedis(request)


AioRedis = Annotated[AsyncRedis, Depends(get_redis_client)]



from __future__ import annotations

import functools
import inspect
import sys
import time
from collections.abc import Awaitable, Callable
from contextlib import (
    AbstractAsyncContextManager,
    AbstractContextManager,
)
from types import TracebackType
from typing import TYPE_CHECKING, Any, TypeVar, overload

if TYPE_CHECKING:  # pragma: no cover
    if sys.version_info >= (3, 11):
        from typing import Self
    else:
        from typing_extensions import Self

T_Retval = TypeVar("T_Retval", Awaitable[Any], Any)


class Timer(AbstractContextManager, AbstractAsyncContextManager):
    """Print time cost of the function.

    Usage::

    ... code-block:: python3

        import time
        from loguru import logger

        def do_staff():
            time.sleep(0.1)

        with Timer('do sth ...'):
            do_staff()
        # do sth ... Cost: 0.1 seconds

        with Timer('do sth ...', verbose=False) as t:
            do_staff()
        assert isinstance(t.cost, float)
        logger.info(str(t))
        # do sth ... Cost: 0.1 seconds

        @Timer
        async def main():
            do_staff()

        await main()
        # main Cost: 0.1 seconds

        @Timer
        def read_text(filename):
            from pathlib import Path
            time.sleep(0.2222)
            return Path(filename).read_text()

        read_text('a.txt')
        # read_text Cost: 0.2 seconds
    """

    def __init__(self, message: str | Callable, decimal_places=1, verbose=True) -> None:
        if callable(message):  # Use as decorator
            func = message
            self.__name__ = message = func.__name__
            self.func: Callable = func
        self.message = message
        self._decimal_places = decimal_places
        self._end = self._start = time.time()
        self._verbose = verbose

    def start(self) -> None:
        self._start = time.time()

    def capture(self, verbose=None) -> None:
        self._end = time.time()
        if verbose is None:
            verbose = self._verbose
        if verbose:
            print(self)

    @property
    def cost(self) -> float:
        return round(self._end - self._start, self._decimal_places)

    def __str__(self) -> str:
        return f"{self.message} Cost: {self.cost} seconds"

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.message!r}, {self._decimal_places}, {self._verbose})"

    async def __aenter__(self) -> Timer:
        return self.__enter__()

    async def __aexit__(self, *args, **kwargs) -> None:
        self.__exit__(*args, **kwargs)

    def __enter__(self) -> Self:
        self.start()
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        if exc_type is not SystemExit or not str(exc_val):
            self.capture()

    def _recreate_cm(self) -> Self:
        return self.__class__(
            getattr(self, "func", None) or self.message,
            self._decimal_places,
            self._verbose,
        )

    def __call__(self, *args, **kwargs) -> Any:
        if (func := getattr(self, "func", None)) is None:
            return None
        if inspect.iscoroutinefunction(func):

            @functools.wraps(func)
            async def inner(*gs, **kw):
                async with self._recreate_cm():
                    return await func(*gs, **kw)

            return inner(*args, **kwargs)
        else:
            with self._recreate_cm():
                return func(*args, **kwargs)


@overload
def timeit(func: str) -> Timer: ...  # pragma: no cover


@overload
def timeit(
    func: Callable[..., T_Retval],
) -> Callable[..., T_Retval]: ...  # pragma: no cover


def timeit(func: str | Callable[..., T_Retval]) -> Timer | Callable[..., T_Retval]:
    """Print time cost of the function.

    Usage::

    ... code-block:: python3

        import anyio, time

        @timeit
        async def main():
            await anyio.sleep(0.1)

        await main()
        # main Cost: 0.1 seconds

        @timeit
        def read_text(filename):
            from pathlib import Path
            time.sleep(0.25)
            return Path(filename).read_text()

        read_text('a.txt')
        # read_text Cost: 0.2 seconds

        args, kwargs = (), {}
        def sync_func(): time.sleep(0.24)
        res = timeit(sync_func)(*args, **kwargs)
        # sync_func Cost: 0.2 seconds

        async def async_func():
            await anyio.sleep(1)
        result = await timeit(async_func)(*args, **kwargs)
        # async_func Cost: 1.0 seconds

        with timeit('message'):
            await async_func()
        # message Cost: 1.0 seconds

    """
    if isinstance(func, str):
        return Timer(func)
    func_name = getattr(func, "__name__", str(func))
    if inspect.iscoroutinefunction(func):

        @functools.wraps(func)
        async def deco(*args, **kwargs) -> T_Retval:
            async with Timer(func_name):
                return await func(*args, **kwargs)

    else:

        @functools.wraps(func)
        def deco(*args, **kwargs) -> T_Retval:
            with Timer(func_name):
                return func(*args, **kwargs)

    return deco



from __future__ import annotations

import functools
import socket
from collections.abc import AsyncGenerator, Callable
from contextlib import AbstractAsyncContextManager, asynccontextmanager
from typing import TYPE_CHECKING, TypeVar

if TYPE_CHECKING:
    from asgi_lifespan import LifespanManager
    from asgi_lifespan._types import ASGIApp
    from fastapi import FastAPI
    from httpx import AsyncClient

T = TypeVar("T")
AsyncClientGenerator = AsyncGenerator["AsyncClient", None]


@asynccontextmanager
async def client_manager(
    app: FastAPI, base_url="http://test", mount_lifespan=True, timeout=30, **kwargs
) -> AsyncClientGenerator:
    """Async test client

    Usage::

    ```py
    from collections.abc import AsyncGenerator

    import pytest
    from asynctor.utils import client_manager
    from httpx import AsyncClient

    from main import app


    @pytest.fixture(scope='session')
    async def client() -> AsyncGenerator[AsyncClient, None]:
        async with client_manager(app) as c:
            yield c


    @pytest.fixture(scope="session")
    def anyio_backend():
        return "asyncio"


    @pytest.mark.anyio
    async def test_api(client: AsyncClient):
        response = await client.get("/")
        assert response.status_code == 200
    ```
    """
    import httpx

    if mount_lifespan:
        from asgi_lifespan import LifespanManager

        async with LifespanManager(app) as manager:
            transport = httpx.ASGITransport(manager.app)
            async with httpx.AsyncClient(
                timeout=timeout, transport=transport, base_url=base_url, **kwargs
            ) as c:
                yield c
    else:
        transport = httpx.ASGITransport(app)
        kwargs.update(transport=transport, base_url=base_url)
        async with httpx.AsyncClient(timeout=timeout, **kwargs) as c:
            yield c


class AsyncTestClient(AbstractAsyncContextManager):
    """Async test client for FastAPI

    :param app: a fastapi instance.
    :param mount_lifespan: if True, auto mount lifespan for app.
    :param base_url: scheme and host for the httpx AsyncClient.
    :param kwargs: other kwargs to pass to httpx AsyncClient.

    Usage::

    ... code-block:: python3

        import pytest
        from asynctor import AsyncTestClient, AsyncClientGenerator
        from httpx import AsyncClient

        from main import app

        @pytest.fixture(scope='session')
        async def client() -> AsyncClientGenerator:
            async with AsyncTestClient(app) as c:
                yield c

        @pytest.fixture(scope="session")
        def anyio_backend():
            return "asyncio"

        @pytest.mark.anyio
        async def test_api(client: AsyncClient):
            response = await client.get("/")
            assert response.status_code == 200

    """

    def __init__(
        self,
        app: FastAPI,
        mount_lifespan=True,
        base_url="http://test",
        timeout=30,
        **kwargs,
    ) -> None:
        self._app = app
        self._mount_lifespan = mount_lifespan
        self._manager: LifespanManager | None = None
        self._client: AsyncClient | None = None
        self._base_url = base_url
        self._timeout = timeout
        self._kwargs = kwargs

    async def __aenter__(self) -> AsyncClient:
        from httpx import ASGITransport, AsyncClient

        app: ASGIApp = self._app
        if self._mount_lifespan:
            from asgi_lifespan import LifespanManager

            self._manager = manager = await LifespanManager(app).__aenter__()
            app = manager.app
        self._client = client = await AsyncClient(
            transport=ASGITransport(app),
            base_url=self._base_url,
            timeout=self._timeout,
            **self._kwargs,
        ).__aenter__()
        return client

    async def __aexit__(self, *args, **kw) -> None:
        if self._client:
            await self._client.__aexit__(*args, **kw)
        if self._manager:
            await self._manager.__aexit__(*args, **kw)


class AttrDict(dict):
    """Support get dict value by attribution

    Usage::
        >>> d = AttrDict({'a': 1, 'b': {'c': 2, 'd': {'e': 3}}})
        >>> d.a == d['a'] == 1
        True
        >>> d.b.c == d['b']['c'] == 2
        True
        >>> d.b.d == d['b']['d'] == {'e': 3}
        True
        >>> d.b.d.e == d['b']['d']['e'] == 3
        True
        >>> dd = AttrDict({'keys': 2}, items=1)
        >>> list(dd.items()) == [('keys', 2), ('items', 1)]
        True
        >>> dd['items'] == 1
        True
        >>> list(dd.keys()) == ['keys', 'items']
        True
        >>> dd['keys'] == 2
        True
        >>> d3 = AttrDict({b'a': 1})
        >>> d3[b'a'] == 1 and getattr(d3, 'a', None) is None
        True
    """

    def __init__(self, *args, **kw) -> None:
        super().__init__(*args, **kw)
        exclude = set(dir(self)) | set(self.__dict__)
        for k, v in self.items():
            if not isinstance(k, str) or k in exclude:
                continue
            if isinstance(v, dict):
                v = self.__class__(v)
            self.__dict__.setdefault(k, v)

    def __str__(self) -> str:
        return super().__repr__()

    def __repr__(self) -> str:
        return self.__class__.__name__ + "(" + super().__repr__() + ")"


def get_machine_ip() -> str:
    r"""Get IP of current machine by socket, if failed, return '127.0.0.1'

    Usage::
        >>> import re
        >>> my_ip = get_machine_ip()
        >>> bool(re.search(r'^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}', my_ip))
        True
        >>> inets = my_ip.split('.')
        >>> len(inets) == 4
        True
        >>> sum(map(lambda x: 0 <= int(x) <= 255, inets))
        4
    """
    with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as s:
        s.settimeout(0)
        try:
            # doesn't even have to be reachable
            s.connect(("10.254.254.254", 1))
            return s.getsockname()[0]
        except Exception:
            return "127.0.0.1"


def cache_attr(func: Callable[..., T]) -> Callable[..., T]:
    """Cache result to class attribute

    Usage::
    ```py
    class A:
        @classmethod
        @cache_attr
        def do_sth(cls):
            ...

    result = A.do_sth()
    ```

    Example::
        >>> import time
        >>> from datetime import datetime
        >>> class A:
        ...     @classmethod
        ...     @cache_attr
        ...     def now(cls) -> datetime:
        ...         return datetime.now()
        >>> now_a = A.now()
        >>> time.sleep(0.1)
        >>> now = datetime.now()
        >>> A.now() == now_a == getattr(A, '-cache-now')
        True
        >>> A.now() < now
        True

    """

    @functools.wraps(func)
    def run(cls) -> T:
        key = "-cache-" + func.__name__
        if hasattr(cls, key):
            return getattr(cls, key)
        res = func(cls)
        setattr(cls, key, res)
        return res

    return run


def _test() -> None:  # pragma: no cover
    import doctest

    doctest.testmod(verbose=True)


if __name__ == "__main__":
    _test()



from __future__ import annotations

import os
from typing import TYPE_CHECKING, Annotated, Any, cast

try:
    from redis.asyncio import Redis
except ImportError:  # pragma: no cover
    from contextlib import AbstractAsyncContextManager

    class Redis(AbstractAsyncContextManager):  # type:ignore[no-redef,assignment]
        def __aexit__(self, *args, **kw): ...


if TYPE_CHECKING:  # pragma: no cover
    import sys

    from fastapi import FastAPI, Request

    if sys.version_info >= (3, 11):
        from typing import Self
    else:
        from typing_extensions import Self


def ensure_redis_is_installed() -> None:
    try:
        import redis  # NOQA:F401
    except ImportError:
        raise RuntimeError(
            "AsyncRedis requires redis to be installed. \n"
            "You can install it with: \n\n"
            "    pip install redis\n"
            "Or:\n"
            '    pip install "asynctor[redis]"\n'
        ) from None


class RedisClient(Redis):
    def __init__(self, **kw: Annotated[Any, "Same as redis.asyncio.Redis"]) -> None:
        """Expand `redis.asyncio.Redis` to auto load REDIS_HOST from os environ,
        if host parameter not explicitly set.

        Example::
            ```
            import os
            os.environ['REDIS_HOST'] = '192.168.0.2'
            r = RedisClient()
            assert r.connection_pool.connection_kwargs['host'] == '192.168.0.2'

            r = RedisClient(host='127.0.0.1')
            assert r.connection_pool.connection_kwargs['host'] == '127.0.0.1'
            ```
        """
        if "host" not in kw and (host := os.getenv("REDIS_HOST")):
            kw["host"] = host
        super().__init__(**kw)


class AsyncRedis(RedisClient):
    """Async redis client for FastAPI

    Usage::
        >>> from contextlib import asynccontextmanager
        >>> from fastapi import FastAPI, Request
        >>> @asynccontextmanager
        ... async def lifespan(fastapi_app: FastAPI):
        ...     async with AsyncRedis(fastapi_app):
        ...         yield
        ...
        >>> app = FastAPI(lifespan=lifespan)
        >>> @app.get('/keys')
        ... async def show_redis_keys(request: Request) -> list[str]:
        ...     redis: AsyncRedis = AsyncRedis(request)
        ...     return await redis.keys()
        ...
        >>> async def use_outside_fastapi():
        ...     async with AsyncRedis(host='localhost') as redis:
        ...         keys: list[str] = await redis.keys()
        ...
    """

    def __new__(
        cls,
        app: FastAPI | Request | Annotated[str, "redis host, e.g.: 127.0.0.1"] | None = None,
        check_connection: bool = True,
        **kw: Annotated[Any, "kwargs that pass to the Redis class"],
    ) -> AsyncRedis:
        """Create a new redis instance or get the redis instance from fastapi state

        :param app: fastapi/fastapi.Request/redis host, if None get redis host from os environ
        :param check_connection: whether check redis host pingable when __aenter__
        :param kw: kwargs that pass to the Redis class when initial
        """
        if (
            app is not None
            and (app := getattr(app, "app", None))
            and (state := getattr(app, "state", None))
        ):
            # app isinstance of fastapi.FastAPI or fastapi.Request
            return cast(AsyncRedis, state.redis)
        return super().__new__(cls)

    def __init__(
        self,
        app: FastAPI | Request | Annotated[str, "redis host, e.g.: 127.0.0.1"] | None = None,
        check_connection: bool = True,
        **kw: Annotated[Any, "kwargs that pass to the Redis class"],
    ) -> None:
        """Create a new redis instance, and then mount to the state if app is a fastapi application

        :param app: fastapi/fastapi.Request/redis host, if None get redis host from os environ
        :param check_connection: whether check redis host pingable when __aenter__
        :param kw: kwargs that pass to the Redis class when initial
        """
        if isinstance(app, str):
            kw.setdefault("host", app)
            app = None
        super().__init__(**kw)
        self._check_connection = check_connection
        if app is not None and hasattr(app, "state"):
            # isinstance(app, FastAPI)
            app.state.redis = self

    async def __aenter__(self) -> Self:
        if not hasattr(self, "aclose"):
            ensure_redis_is_installed()
        # Check connection when app startup
        if self._check_connection:
            await self.ping()
        return self
